/*
 * Copyright  2001-2005
 * University of Maryland
 * All Rights Reserved
 * 
 * Written by Ben Bederson, Aaron Clamage
 * bederson@cs.umd.edu
 * http://www.cs.umd.edu/~bederson
 * 
 * RangeSlider - a slider with two "heads" that defines a model with
 * "low" and "high" values, constrained by the slider's minimum and maximum.
 */

using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;
using System.Threading;

namespace UMD.HCIL {	
	public delegate void ScrollEventHandler(object sender, ScrollEventArgs e);

	public class RangeSlider : System.Windows.Forms.UserControl {
		public event ScrollEventHandler Scroll;

		#region Internal Variables
		static private Brush lightGrayBrush = new SolidBrush(SystemColors.Control);
		static private Brush TROUGH_BRUSH = new SolidBrush(Color.FromArgb(220, 220, 220));
		static private int DEFAULT_MINIMUM = -8;
		static private int DEFAULT_MAXIMUM = 52;
		static private int DEFAULT_LOW = 0;
		static private int DEFAULT_HIGH = 11;
		static private int MIN_THUMB_LENGTH = 0;

		private System.ComponentModel.Container components = null;
		private int lowTroughPageLength = 4;
		private int highTroughPageLength = 4;
		private ButtonState minimumButtonState = ButtonState.Normal;
		private ButtonState maximumButtonState = ButtonState.Normal;
		private ButtonState lowButtonState = ButtonState.Normal;
		private ButtonState highButtonState = ButtonState.Normal;
		private bool lowTroughPushed = false;
		private bool highTroughPushed = false;
		private Brush markBrush = new SolidBrush(Color.Yellow);
		private Pen markPen = new Pen(Color.Black);
		private Brush troughBrush = TROUGH_BRUSH;
		private Brush transparentMarkBrush = new SolidBrush(
			Color.FromArgb((int)Utils.Lerp((float).8, (float)255, (float)SystemColors.Control.G),
			(int)Utils.Lerp((float).8, (float)255, (float)SystemColors.Control.R), 
			(int)Utils.Lerp((float).8, (float)0, (float)SystemColors.Control.B)));		
		private Pen transparentMarkPen = new Pen(
			Color.FromArgb((int)Utils.Lerp((float).8, (float)0, (float)SystemColors.Control.G),
			(int)Utils.Lerp((float).8, (float)0, (float)SystemColors.Control.R), 
			(int)Utils.Lerp((float).8, (float)0, (float)SystemColors.Control.B)));						
		private int minimum = DEFAULT_MINIMUM;
		private int maximum = DEFAULT_MAXIMUM;
		private int low = DEFAULT_LOW;
		private int high = DEFAULT_HIGH;
		private bool thumbPressed = false;
		private float dy = 0;
		private int prevY = -1;
		private System.Windows.Forms.Timer buttonTimer = null;
		private int layoutHeight;
		private int lowStart;
		private int thumbStart;
		private int thumbEnd;
		private int highStart;
		private int yOffset;
		private ArrayList marks;
		#endregion

		/**********************************************************************
		 * PUBLIC API
		 *********************************************************************/
		public RangeSlider() : this(0, 100, 0, 10) {
		}

		public RangeSlider(int min, int max, int low, int high) {
			InitializeComponent();

			this.minimum = min;
			this.maximum = max;
			this.low = low;
			this.high = high;

			MouseMove += new MouseEventHandler(mouseMoveEventHandler);
			Resize += new EventHandler(resizeEventHandler);
			CalculatePositions();
		}

		public int Minimum {
			get { return minimum; }
			set {
				minimum = value;
				if (minimum > low) {
					minimum = low;
				}
				CalculatePositions();
				Invalidate();
			}
		}

		public int Maximum {
			get { return maximum; }
			set {
				maximum = value;
				if (maximum < high) {
					maximum = high;
				}
				CalculatePositions();
				Invalidate();
			}
		}

		public int Low {
			get { return low; }
			set {
				Rectangle oldRect = new Rectangle(0, lowStart-10, Width, 2 * Width + (thumbEnd - thumbStart)+20);

				low = value;
				if (low < minimum) low = minimum;
				if (low > high) low = high;
				CalculatePositions();

				Rectangle newRect = new Rectangle(0, lowStart-10, Width, 2 * Width + (thumbEnd - thumbStart)+20);

				Invalidate(Utils.RectangleUnion(oldRect, newRect));
				Update();
			}
		}

		public int High {
			get { return high; }
			set {
				Rectangle oldRect = new Rectangle(0, lowStart-10, Width, 2 * Width + (thumbEnd - thumbStart)+20);

				high = value;
				if (high < low) {
					high = low;
				}
				if (high > maximum) {
					high = maximum;
				}

				CalculatePositions();

				Rectangle newRect = new Rectangle(0, lowStart-10, Width, 2 * Width + (thumbEnd - thumbStart)+20);

				Invalidate(Utils.RectangleUnion(oldRect, newRect));
				Update();
			}
		}

		public int LowTroughPageLength {
			get { return lowTroughPageLength; }
			set { lowTroughPageLength = value; }
		}

		public int HighTroughPageLength {
			get { return highTroughPageLength; }
			set { highTroughPageLength = value; }
		}

		/// <summary>
		/// An array of integers that correspond to values within the rangeslider's min/max range.
		/// Each integer will be drawn on the slider as a rectangular "mark" - even if it is outside
		/// the current low/high range.
		/// </summary>
		public ArrayList Marks {
			get { return marks; }
			set { 
				marks = value;
				Invalidate(new Rectangle(0, Width, Width, Height - 2 * Width));
			}
		}

		/**********************************************************************
		 * INTERNAL IMPLEMENTATION
		 *********************************************************************/

		#region Component Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify 
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent() {
			// 
			// RangeSlider
			//
			//this.Name = "RangeSlider";
			this.Size = new System.Drawing.Size(16, 232);

		}
		#endregion

		#region Internal Implementation
		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing ) {
			if( disposing ) {
				if( components != null )
					components.Dispose();
			}
			base.Dispose( disposing );
		}

		private int HighStart {
			get { return highStart; }
		}

		private int LowStart {
			get { return lowStart; }
		}

		private System.Windows.Forms.Timer Timer {
			get {
				if (buttonTimer == null) {
					buttonTimer = new System.Windows.Forms.Timer();
					buttonTimer.Tick += new EventHandler(UpdateTimerTick);
					buttonTimer.Enabled = false;
					buttonTimer.Interval = 100;
				}
				return buttonTimer;
			}
		}

		private void SetValues(int minimum, int maximum, int low, int high) {
			if (minimum <= low && low <= high && high <= maximum) {
				this.minimum = minimum;
				this.maximum = maximum;
				this.low = low;
				this.high = high;
			}
			CalculatePositions();
			Invalidate();
		}

		private void OnScroll(ScrollEventArgs e) {
			if (Scroll != null) {
				Scroll(this, e);
			}
		}

		private void MoveThumb(int units, bool bounded) {
			IncrementThumb(units, bounded);
			FireScroll();
			OnScroll(new ScrollEventArgs(ScrollEventType.EndScroll, 0));
		}

		private void IncrementThumb(int dt, bool bounded) {
			Rectangle oldRect = new Rectangle(0, lowStart-10, Width, 2 * Width + (thumbEnd - thumbStart)+20);
			int oldThumbStart = thumbStart;
			int oldThumbEnd = thumbEnd;
			int y, height;

			low += dt;
			high += dt;

			if (bounded) {
				if (low < minimum) {
					int dm = minimum - low;
					low += dm;
					high += dm;
				}
				if (high > maximum) {
					int dm = high - maximum;
					low -= dm;
					high -= dm;
				}
			} 
			else {
				if (low < minimum) {
					int dm = minimum - low;
					minimum -= dm;
					maximum -= dm;
				}
				if (high > maximum) {
					int dm = high - maximum;
					minimum += dm;
					maximum += dm;
				}
			}

			CalculatePositions();

			if (dt < 0) {
				y = thumbStart;
				height = oldThumbEnd - thumbStart;
			} 
			else {
				y = oldThumbStart;
				height = thumbEnd - oldThumbStart;
			}
			Rectangle newRect = new Rectangle(0, lowStart-10, Width, 2 * Width + (thumbEnd - thumbStart)+20);

			Invalidate(Utils.RectangleUnion(oldRect, newRect));
			Update();
		}

		private void CalculatePositions() {
			int offset = Width;
			layoutHeight = Height - 2 * Width;

			offset = 2 * Width;
			layoutHeight = Height - (4 * Width);

			thumbStart = (int)(offset + (float)(Low - Minimum) / (Maximum - Minimum) * layoutHeight);
			thumbEnd = (int)(offset + (float)(High - Minimum) / (Maximum - Minimum) * layoutHeight);
			if (thumbEnd < (thumbStart + MIN_THUMB_LENGTH)) {
				thumbEnd = thumbStart + MIN_THUMB_LENGTH;
				if (thumbEnd > (Height - offset)) {
					thumbEnd = Height - offset;
					thumbStart = thumbEnd - MIN_THUMB_LENGTH;
				}
			}

			lowStart = thumbStart - Width;
			highStart = thumbEnd;
		}

		private void MinimumFired() {
			if (Low == Minimum) {
				minimum--;
				maximum--;
			}
			IncrementThumb(-1, true);
			OnScroll(new ScrollEventArgs(ScrollEventType.ThumbTrack, 0));
		}

		private void MaximumFired() {
			if (High == Maximum) {
				minimum++;
				maximum++;
			}
			IncrementThumb(1, true);
			OnScroll(new ScrollEventArgs(ScrollEventType.ThumbTrack, 0));
		}

		private void LowFired() {
			if (Low > Minimum) {
				Low--;
			}
		}

		private void HighFired() {
			if (High < Maximum) {
				High++;
			}
		}

		private void LowTroughFired() {
			if (Low > (Minimum + lowTroughPageLength)) {
				IncrementThumb(- lowTroughPageLength, true);
			} 
			else {
				int delta = Low - Minimum;
				IncrementThumb(- delta, true);
			}
			OnScroll(new ScrollEventArgs(ScrollEventType.ThumbTrack, 0));
		}

		private void HighTroughFired() {
			if (High < (Maximum - highTroughPageLength)) {
				IncrementThumb(highTroughPageLength, true);
			} 
			else {
				int delta = Maximum - High;
				IncrementThumb(delta, true);
			}
			OnScroll(new ScrollEventArgs(ScrollEventType.ThumbTrack, 0));
		}

		private void resizeEventHandler(object sender, EventArgs e) {
			CalculatePositions();
			Invalidate();
		}

		protected override void OnMouseDown(MouseEventArgs e) {
			base.OnMouseDown(e);
			dy = 0;

			if (e.Y < Width) {
				// Minimum button
				minimumButtonState = ButtonState.Pushed;
				Invalidate(new Rectangle(0, 0, Width, Width));
				Update();
				MinimumFired();
				Timer.Enabled = true;
				Invalidate();
			} 
			else if (e.Y < lowStart) {
				// Low trough
				LowTroughFired();
				lowTroughPushed = true;
				Timer.Enabled = true;
				Invalidate();
			} 
			else if (e.Y < thumbStart) {
				// Low button
				lowButtonState = ButtonState.Pushed;
				prevY = e.Y;
				yOffset = prevY - lowStart;
				Invalidate(new Rectangle(0, lowStart, Width, Width));
				Update();
			} 
			else if (e.Y < thumbEnd) {
				// Thumb 
				thumbPressed = true;
				prevY = e.Y;
				yOffset = prevY - thumbStart;
			} 
			else if (e.Y < (highStart + Width)) {
				// High button
				highButtonState = ButtonState.Pushed;
				prevY = e.Y;
				yOffset = prevY - highStart;
				Invalidate(new Rectangle(0, highStart, Width, Width));
				Update();
			} 
			else if (e.Y < (Height - Width)) {
				// High trough
				HighTroughFired();
				highTroughPushed = true;
				Timer.Enabled = true;
				Invalidate();
			} 
			else {
				// Maximum button
				maximumButtonState = ButtonState.Pushed;
				Invalidate(new Rectangle(0, Height-Width, Width, Width));
				Update();
				MaximumFired();
				Timer.Enabled = true;
			}
		}

		private void mouseMoveEventHandler(object sender, MouseEventArgs e) {
			bool moved = false;
			bool limit = false;
			int startY = 0;
			int idy = 0;

			dy += (float)(e.Y - prevY) / layoutHeight * (Maximum - Minimum);

			if (thumbPressed) {
				if ((Low + dy) < Minimum) {
					dy = Minimum - Low;
				}
				if ((High + dy) > Maximum) {
					dy = Maximum - High;
				}
				idy = (int)Math.Round(dy);
				if (idy != 0) {
					IncrementThumb(idy, true);
					moved = true;
				}

				startY = thumbStart + yOffset;
				if (lowStart == Width || highStart == Height - (2 *Width)) limit = true;
			} 
			else if (lowButtonState == ButtonState.Pushed) {
				if ((Low + dy) <= Minimum) {
					dy = Minimum - Low;
				}
				idy = (int)Math.Round(dy);
				if (idy != 0) {
					Low += idy;
					moved = true;
				}

				startY = lowStart + yOffset;
				
				if ((lowStart == Width) && (e.Y < startY) ||
					(high - low) == 0 && e.Y > startY)
					limit = true;
			} 
			else if (highButtonState == ButtonState.Pushed) {
				if ((High + dy) >= Maximum) {
					dy = Maximum - High;
				}
				idy = (int)Math.Round(dy);
				if (idy != 0) {
					High += idy;
					moved = true;
				}

				startY = highStart + yOffset;
				if ((highStart == Height - (2 * Width)) && (e.Y > startY) ||
					(high - low) == 0 && e.Y < startY)
					limit = true;
			}

			dy -= idy;
			prevY = e.Y;

			if (moved) {
				FireScroll();
			}

			if (limit) dy += (float)(prevY - startY) / layoutHeight * (Maximum - Minimum);
		}

		private void FireScroll() {
			ScrollEventType scrollType = ScrollEventType.ThumbTrack;
			OnScroll(new ScrollEventArgs(scrollType, 0));
		}

		protected override void OnMouseUp(MouseEventArgs e) {
			if (buttonTimer != null) {
				buttonTimer.Enabled = false;
				buttonTimer.Dispose();
				buttonTimer = null;
			}

			if ((minimumButtonState == ButtonState.Pushed)) {
				// Minimum button
				minimumButtonState = ButtonState.Normal;
				Invalidate(new Rectangle(0, 0, Width, Width));
				Update();
			} 
			else if ((lowButtonState == ButtonState.Pushed)) {
				// Low button - do nothing
				lowButtonState = ButtonState.Normal;
				Invalidate(new Rectangle(0, lowStart, Width, Width));
				Update();
			} 
			else if ((highButtonState == ButtonState.Pushed)) {
				// High button - do nothing
				highButtonState = ButtonState.Normal;
				Invalidate(new Rectangle(0, highStart, Width, Width));
				Update();
			} 
			else if (maximumButtonState == ButtonState.Pushed) {
				// Maximum button
				maximumButtonState = ButtonState.Normal;
				Invalidate(new Rectangle(0, Height-Width, Width, Width));
				Update();
			}

			minimumButtonState = ButtonState.Normal;
			maximumButtonState = ButtonState.Normal;
			lowButtonState = ButtonState.Normal;
			highButtonState = ButtonState.Normal;
			lowTroughPushed = false;
			highTroughPushed = false;
			thumbPressed = false;

			OnScroll(new ScrollEventArgs(ScrollEventType.EndScroll, 0));
			base.OnMouseUp(e);
		}


		private void UpdateTimerTick(object sender, System.EventArgs e) {
			if (minimumButtonState == ButtonState.Pushed) {
				MinimumFired();
			} 
			else if (maximumButtonState == ButtonState.Pushed) {
				MaximumFired();
			}
			else if (lowTroughPushed) {
				LowTroughFired();
			}
			else if (highTroughPushed) {
				HighTroughFired();
			}
		}

		protected override void OnPaint(PaintEventArgs pe) {
			Graphics g = pe.Graphics;
			Rectangle clipRectangle = pe.ClipRectangle;
			Rectangle thumbRect;

			thumbRect = new Rectangle(0, lowStart, Width, 2 * Width + (thumbEnd - thumbStart));

			// trough
			Rectangle troughRect = new Rectangle(0, Width, Width, Height - 2 * Width);
			if (clipRectangle.IntersectsWith(troughRect)) g.FillRectangle(troughBrush, troughRect);

			// draw solid marks under the thumb
			if (clipRectangle.IntersectsWith(troughRect) && (marks != null)) {
				int markStart;
				float fmarkLength = (float)(Height - 4*Width - 5) / (Maximum - Minimum);
				int markLength = (int)fmarkLength;
				int width = Width - 4;
				foreach (int mark in marks) {
					markStart = (int)(2*Width + (mark - Minimum) * fmarkLength);
					Rectangle markRect = new Rectangle(2, markStart, width, markLength+1);
					if (clipRectangle.IntersectsWith(markRect)) {
						if ((markStart+markLength < lowStart) || ((markStart) > (highStart + Width - 2))) {
							//if (((markStart+markLength) > 2*Width)
							//	&& (markStart+markLength < Height-2*Width)
							//	&& (((markStart + markLength) < lowStart) || ((markStart) > (highStart + Width - 2))))
							//{
							g.FillRectangle(markBrush, 2, markStart, width, markLength);
							g.DrawRectangle(markPen, 2, markStart, width, markLength);
						}
					}
				}
			}

			// thumb background
			if (clipRectangle.IntersectsWith(thumbRect) && thumbEnd >= thumbStart) {
				g.FillRectangle(lightGrayBrush, 0, lowStart, Width, 2 * Width + (thumbEnd - thumbStart));
			}

			// draw transparent marks over the thumb
			if (clipRectangle.IntersectsWith(thumbRect) && (marks != null)) {
				int markStart;
				float fmarkLength = (float)(Height - 4*Width - 5) / (Maximum - Minimum);
				int markLength = (int)fmarkLength;
				int width = Width - 6;
				foreach (int mark in marks) {
					markStart = (int)(2*Width + (mark - Minimum) * fmarkLength);
					Rectangle markRect = new Rectangle(2, markStart, width, markLength);
					if (clipRectangle.IntersectsWith(markRect)) {
						if (((markStart + markLength) >= lowStart) && ((markStart) <= (highStart + Width - 2))) {
							g.FillRectangle(transparentMarkBrush, 2, markStart, width, markLength);
							g.DrawRectangle(transparentMarkPen, 2, markStart, width, markLength);
						}
					}
				}
			}

			// thumb
			if (clipRectangle.IntersectsWith(thumbRect) && thumbEnd > thumbStart) {
				ControlPaint.DrawButton(g, 0, thumbStart, Width, thumbEnd - thumbStart, ButtonState.Normal, true);
			}

			// low button
			Rectangle lowRect = new Rectangle(0, lowStart, Width, Width);
			if (clipRectangle.IntersectsWith(lowRect)) {
				ControlPaint.DrawScrollButton(g, 0, lowStart, Width, Width, ScrollButton.Down, lowButtonState, true);
			}

			// high button
			Rectangle highRect = new Rectangle(0, highStart, Width, Width);
			if (clipRectangle.IntersectsWith(highRect)) {
				ControlPaint.DrawScrollButton(g, 0, highStart, Width, Width, ScrollButton.Up, highButtonState, true);
			}

			// minimum button
			Rectangle minRect = new Rectangle(0, 0, Width, Width);
			if (clipRectangle.IntersectsWith(minRect)) {
				ControlPaint.DrawScrollButton(g, 0, 0, Width, Width, ScrollButton.Up, minimumButtonState, false);
			}

			// maximum button
			Rectangle maxRect = new Rectangle(0, Height-Width, Width, Width);
			if (clipRectangle.IntersectsWith(maxRect)) {
				ControlPaint.DrawScrollButton(g, 0, Height-Width, Width, Width, ScrollButton.Down, maximumButtonState, false);			
			}
		}
		#endregion
	}
}